/*-
 * Copyright (C) @BABOLO  2002 Oct 05
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2002 Oct 05\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: ip_extend.c,v 1.2 2004/10/16 21:56:49 babolo Exp $";
#endif /* not lint */

#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <postgres.h>
#include <funcapi.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <fmgr.h>
#include <utils/inet.h>
#include <nodes/nodes.h>
#include <nodes/execnodes.h>
#define ip_bits(inetptr)     (((inet_struct *)VARDATA(inetptr))->bits)
#define ip_family(inetptr)   (((inet_struct *)VARDATA(inetptr))->family)
#define ip_type(inetptr)     (((inet_struct *)VARDATA(inetptr))->type)
#define ip_addr(inetptr)     (((inet_struct *)VARDATA(inetptr))->ip_addr)
#define ip_maxbits(inetptr)  (ip_family(inetptr) == PGSQL_AF_INET ? 32 : 128)
#define INET_MEMSIZE(a) (VARHDRSZ + ((char*)ip_addr(a) - (char*)VARDATA(a)) + ip_addrsize(a))
#define fctx ((results*)funcctx->user_fctx)

typedef struct {
    u_long current;
    u_long brodcast;
    char   done;
} results;

static int
ip_addrsize(inet *inetptr) {
    switch (ip_family(inetptr)) {
    case PGSQL_AF_INET:  return 4;
    case PGSQL_AF_INET6: return 16;
    default:             return -1;
}   }


u_long
ip_v4addr(inet *inetptr) {
    int i;
    u_long a;
    u_char *p;

    for (p = ip_addr(inetptr), a = 0, i = 0; i < ip_addrsize(inetptr); i++) {
        a = (a << 8) | p[i];
    }
    return(a);
}

void
toip_v4addr(inet *inetptr, u_long a) {
    int i;
    u_char *p;

    for (p = ip_addr(inetptr), i = ip_addrsize(inetptr) - 1; i >= 0; i--) {
        p[i] = a;
        a >>= 8;
}   }

PG_FUNCTION_INFO_V1(ip_set);
Datum
ip_set(PG_FUNCTION_ARGS) {
    u_long mask, step;
    int maxbits, nowbits;
    inet *src, *dst;
    FuncCallContext  *funcctx;
    MemoryContext oldcontext;

    if  (PG_ARGISNULL(0) || PG_ARGISNULL(1)) {
        PG_RETURN_NULL();
    }
    src = PG_GETARG_INET_P(0);
    if  (ip_family(src) == PGSQL_AF_INET6)
        ereport(ERROR, ( errcode(ERRCODE_FEATURE_NOT_SUPPORTED)
                       , errmsg("ip_set: address family AF_INET6 not supported yet")
               )       );
    if  (ip_family(src) != PGSQL_AF_INET)
                                    elog(ERROR, "ip_set: unknown address family (%d)", ip_family(src));
    maxbits = ip_maxbits(src);
    nowbits = PG_GETARG_INT32(1);
    if  (nowbits > maxbits || nowbits < 0)
        ereport(ERROR, ( errcode(ERRCODE_INVALID_PARAMETER_VALUE)
                       , errmsg("ip_set: Out of range")
                       , errdetail("Parameter $2 = %d out of range (0..%d)", nowbits, maxbits)
               )       );
    if  (ip_bits(src) > nowbits) {
        PG_RETURN_NULL();
    }
    step = 1 << (maxbits - nowbits);
    if  (SRF_IS_FIRSTCALL()) {
        funcctx = SRF_FIRSTCALL_INIT();
        mask = 0xFFFFFFFF << (maxbits - ip_bits(src));
        oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
        fctx = (results*)palloc(sizeof(results));
        if  (fctx == NULL)
                     ereport(ERROR, (errcode(ERRCODE_OUT_OF_MEMORY), errmsg("ip_set: out of memory")));
        MemSet(fctx, 0, sizeof(results));
        funcctx->max_calls = 1 << (nowbits - ip_bits(src));           /* BEWARE!!! Overflow possible */
        fctx->current = mask & ip_v4addr(src);
        fctx->brodcast = ~mask | ip_v4addr(src);
        fctx->done = 0;
        MemoryContextSwitchTo(oldcontext);
    }
    funcctx = SRF_PERCALL_SETUP();
    if  (fctx->current <= fctx->brodcast) {
        dst = (inet*)palloc(VARHDRSZ + sizeof(inet_struct));
        if  (dst == NULL)
                     ereport(ERROR, (errcode(ERRCODE_OUT_OF_MEMORY), errmsg("ip_set: out of memory")));
        MemSet(dst, 0, VARHDRSZ + sizeof(inet_struct));
        ip_family(dst) = PGSQL_AF_INET;
        ip_bits(dst) = nowbits;
        ip_type(dst) = 0;
        toip_v4addr(dst, fctx->current);
        VARATT_SIZEP(dst) = INET_MEMSIZE(dst);
        fctx->current += step;
        SRF_RETURN_NEXT(funcctx, InetPGetDatum(dst));
    }else{
        SRF_RETURN_DONE(funcctx);
}   }

PG_FUNCTION_INFO_V1(ip_supernet);
Datum
ip_supernet(PG_FUNCTION_ARGS) {
    inet *src, *dst;
    FuncCallContext  *funcctx;
    MemoryContext oldcontext;


    if  (PG_ARGISNULL(0)) {
        PG_RETURN_NULL();
    }

    src = PG_GETARG_INET_P(0);
    if  (ip_family(src) == PGSQL_AF_INET6)
        ereport(ERROR, ( errcode(ERRCODE_FEATURE_NOT_SUPPORTED)
                       , errmsg("ip_supernet: address family AF_INET6 not supported yet")
               )       );
    if  (ip_family(src) != PGSQL_AF_INET)
                               elog(ERROR, "ip_supernet: unknown address family (%d)", ip_family(src));
    if  (SRF_IS_FIRSTCALL()) {
        funcctx = SRF_FIRSTCALL_INIT();
        oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
        fctx = (results*)palloc(sizeof(results));
        if  (fctx == NULL)
                ereport(ERROR, (errcode(ERRCODE_OUT_OF_MEMORY), errmsg("ip_supernet: out of memory")));
        MemSet(fctx, 0, sizeof(results));
        funcctx->max_calls = fctx->done = ip_bits(src);
        fctx->current = ip_v4addr(src);
        MemoryContextSwitchTo(oldcontext);
    }
    funcctx = SRF_PERCALL_SETUP();
    if  (--(fctx->done) >= 0) {
        dst = (inet*)palloc(VARHDRSZ + sizeof(inet_struct));
        if  (dst == NULL)
                ereport(ERROR, (errcode(ERRCODE_OUT_OF_MEMORY), errmsg("ip_supernet: out of memory")));
        MemSet(dst, 0, VARHDRSZ + sizeof(inet_struct));
        ip_family(dst) = PGSQL_AF_INET;
        ip_bits(dst) = fctx->done;
        ip_type(dst) = 0;
        toip_v4addr(dst, fctx->current);
        VARATT_SIZEP(dst) = INET_MEMSIZE(dst);
        SRF_RETURN_NEXT(funcctx, InetPGetDatum(dst));
    }else{
        SRF_RETURN_DONE(funcctx);
}   }

PG_FUNCTION_INFO_V1(int4_inet);
Datum
int4_inet(PG_FUNCTION_ARGS) {
    inet *dst;

    if  (PG_ARGISNULL(0)) PG_RETURN_NULL();
    dst = (inet*)palloc(VARHDRSZ + sizeof(inet_struct));
    if  (dst == NULL)
                  ereport(ERROR, (errcode(ERRCODE_OUT_OF_MEMORY), errmsg("int4_inet: out of memory")));
    MemSet(dst, 0, VARHDRSZ + sizeof(inet_struct));
    ip_family(dst) = PGSQL_AF_INET;
    ip_bits(dst) = 32;
    ip_type(dst) = 0;
    toip_v4addr(dst, PG_GETARG_INT32(0));
    VARATT_SIZEP(dst) = INET_MEMSIZE(dst);
    PG_RETURN_INET_P(dst);
};

PG_FUNCTION_INFO_V1(int8_inet);
Datum
int8_inet(PG_FUNCTION_ARGS) {
    inet *dst;

    if  (PG_ARGISNULL(0)) PG_RETURN_NULL();
    dst = (inet*)palloc(VARHDRSZ + sizeof(inet_struct));
    if  (dst == NULL)
                  ereport(ERROR, (errcode(ERRCODE_OUT_OF_MEMORY), errmsg("int8_inet: out of memory")));
    MemSet(dst, 0, VARHDRSZ + sizeof(inet_struct));
    ip_family(dst) = PGSQL_AF_INET;
    ip_bits(dst) = 32;
    ip_type(dst) = 0;
    toip_v4addr(dst, (u_long)PG_GETARG_INT64(0));
    VARATT_SIZEP(dst) = INET_MEMSIZE(dst);
    PG_RETURN_INET_P(dst);
};

PG_FUNCTION_INFO_V1(inet_int4);
Datum
inet_int4(PG_FUNCTION_ARGS) {
    inet *src;

    if  (PG_ARGISNULL(0)) PG_RETURN_NULL();
    src = PG_GETARG_INET_P(0);
        switch (ip_family(src)) {
    case PGSQL_AF_INET:  PG_RETURN_INT32(ip_v4addr(src));
    case PGSQL_AF_INET6:
        ereport(ERROR, ( errcode(ERRCODE_FEATURE_NOT_SUPPORTED)
                       , errmsg("address AF_INET6 too big for int4")
               )       );
    default:
        elog(ERROR, "unknown address family (%d)", ip_family(src));
    }
    PG_RETURN_NULL();
}

PG_FUNCTION_INFO_V1(inet_int8);
Datum
inet_int8(PG_FUNCTION_ARGS) {
    inet *src;

    if  (PG_ARGISNULL(0)) PG_RETURN_NULL();
    src = PG_GETARG_INET_P(0);
    switch (ip_family(src)) {
    case PGSQL_AF_INET:  PG_RETURN_INT64(ip_v4addr(src));
    case PGSQL_AF_INET6:
        ereport(ERROR, ( errcode(ERRCODE_FEATURE_NOT_SUPPORTED)
                       , errmsg("address AF_INET6 too big for int8")
               )       );
    default:
        elog(ERROR, "unknown address family (%d)", ip_family(src));
    }
    PG_RETURN_NULL();
}

PG_FUNCTION_INFO_V1(inet_type);
Datum
inet_type(PG_FUNCTION_ARGS) {
    inet *src;

    if  (PG_ARGISNULL(0)) PG_RETURN_NULL();
    src = PG_GETARG_INET_P(0);
    if  (ip_family(src) == PGSQL_AF_INET || ip_family(src) == PGSQL_AF_INET6) {
        PG_RETURN_INT16(ip_type(src));
    }else{
        elog(ERROR, "unknown address family (%d)", ip_family(src));
        PG_RETURN_NULL();
}   }
